/*
	audio_functions: audio data manipulation

	Note that only fullchansample is actually used (and debugged), the others are more of historic importance.

	part of DerMixD
	(c)2004-8 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"

#if 0 // Still... the code stays here for reference.
/*
	fullsample()
	makes the a samples of A to b samples of B and multiplies by vol on the way.
	
	in other terms: resample and amplify
	
	It just picks single input samples on more or less appropriate positions without any filtering
*/

void fullsample(audio_type* A, unsigned int a, int* B, unsigned int b, double vol)
{
	//cout << "fullsample got inbuffer at " << A << " and outbuffer at " << B << endl;
	if(vol == 0)
	{
		for(unsigned int i = 0; i < 2*b; ++i) B[i] = 0;
	}
	else
	{
		if(a == b)
		{
			if(vol == 1)
			{
				for(unsigned int i = 0; i < 2*b; ++i) B[i] = A[i];
			}
			else
			{		
				for(unsigned int i = 0; i < 2*b; ++i)	B[i] = (int) (vol*A[i]);
			}
		}
		else
		{
			double r = 1.0*a/b;
			if(vol == 1)
			{
				for(unsigned int i = 0; i < b; ++i)
				{
						unsigned int index = 2* (unsigned int) (r*i);
						B[2*i] = A[index]; //gerade
						B[2*i+1] = A[index + 1]; //ungerade
				}
			}
			else
			{
				for(unsigned int i = 0; i < b; ++i)
				{
						unsigned int index = 2* (unsigned int) (r*i);
						B[2*i] = (int) (vol * A[index]); //gerade
						B[2*i+1] = (int) (vol * A[index + 1]); //ungerade
				}
			}
		}
	}
}

/*
	fullsample_hq()
	makes the a samples of A to b samples of B and multiplies by vol on the way, hopefully a bit more pleasing to the ear (but presumably way slower) than the variant above.
	
	in other terms: resample and amplify
	
	More quality means: All insamples have influence. When downsampling (meaning speedup), the mean value of some insamples make one outsample. Simple. 
	But how to improve slowdown / upsampling? The simple method of repeating one input frame several times produces nasty clicks, obviously. A first approach would be to make a linear fade:
	
	X                               Y
	x..0.9*(x-y)+y..0.5*(x-y)+y...  y
	
	problem: I need border values for not having ticks from chunk to chunk. That said: this function should return its last calculated value and accept the last value of the last run as additional parameter (array with two elements)... this holds the new endvalues after function run.
	
	This is how it works, then...
*/

void fullsample_hq(audio_type* A, unsigned int a, int* B, unsigned int b, double vol, int* startval)
{
	if(a*b != 0) //should be some kind of error... or not?
	{
		//cout << "fullsample got inbuffer at " << A << " (" << a << " samples) and outbuffer at " << B << " (" << b << " samples)" << endl;
		if(vol == 0)
		{
			for(unsigned int i = 0; i < 2*b; ++i) B[i] = 0;
		}
		else
		{
			if(a == b) for(unsigned int i = 0; i < 2*b; ++i) B[i] = (int) (vol*A[i]);
			else
			{
				float r = 1.0*a/b;
				if(r > 1) //a > b ... downsample, arithmetic mean
				{
					for(unsigned int i = 0; i < b; ++i)
					{
						float sum = 0;
						float sum2 = 0;
						unsigned int n = 0;
						for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
						{
							++n;
							sum += A[2*j];
							sum2 += A[2*j+1];
						}
						B[2*i] = (int) (vol*sum/n); //gerade
						B[2*i+1] = (int) (vol*sum2/n); //ungerade
					}
				}
				else //(a < b)... linear interpolation
				{
					
					for(unsigned int i = 1; i <= b; ++i)
					{
						//i-1 is the sample index!!!
						float part = r*i;  //0.67 ... 1 means the first "present" input sample (A[0] and A[1])
						int base = (int) part; // -1: history, 0 - A[0]
						part = part - base; // tending with 0.67 to 1 (the first input sample)
						--base;
						B[2*(i-1)] = (int)
							(
								(base > -1) ?
								(vol*(part*((int)A[2*(base+1)] - A[2*base]) + A[2*base])) :
								(part*(vol*A[0] - startval[0])+ startval[0])
							); //gerade
						B[2*i-1] = (int)
							(
								(base > -1) ?
								(vol*(part*((int)A[2*(base+1)+1] - A[2*base+1]) + A[2*base+1])) :
								(part*(vol*A[1] - startval[1]) + startval[1])
							); //ungerade
					}
					
				}
			}
		}
		startval[0] = B[2*b-2]; startval[1] = B[2*b-1];
	}
}

//flexible with channels, but slow... (factor 2)
void fullsample_hqbuff(audio_buffer& A, int* B, unsigned int b, double vol, int* startval)
{
	if(A.fill*b != 0) //should be some kind of error... or not?
	{
		//cout << "fullsample got inbuffer at " << A << " (" << a << " samples) and outbuffer at " << B << " (" << b << " samples)" << endl;
		if(vol == 0)
		{
			for(unsigned int i = 0; i < 2*b; ++i) B[i] = 0;
		}
		else
		{
			if(A.fill == b) for(unsigned int i = 0; i < A.channels*b; ++i) B[i] = (int) (vol*A.ptr[i]);
			else
			{
				float r = 1.0*A.fill/b;
				if(r > 1) //median
				{
					float* sum = new float[A.channels];
					for(unsigned int i = 0; i < b; ++i)
					{
						unsigned int n = 0;
						for(unsigned char k = 0; k < A.channels; ++k) sum[k] = 0;
						for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
						{
							++n;
							for(unsigned char k = 0; k < A.channels; ++k) sum[k] += A.ptr[A.channels*j+k];
						}
						for(unsigned char k = 0; k < A.channels; ++k) B[A.channels*i+k] = (int) (vol*sum[k]/n); 
					}
					delete sum;
				}
				else //(a < b)... interpolate
				{
					
					for(unsigned int i = 1; i <= b; ++i)
					{
						//i-1 is the sample index!!!
						float part = r*i;  //0.67 ... 1 means the first "present" input sample (A[0] and A[1])
						int base = (int) part; // -1: history, 0 - A[0]
						part = part - base; // tending with 0.67 to 1 (the first input sample)
						--base;
						
						for(unsigned char k = 0; k < A.channels; ++k) B[A.channels*(i-1)+k] = (int) ((base > -1) ? (vol*(A.ptr[A.channels*base+k] + part*(A.ptr[A.channels*(base+1)+k] - A.ptr[A.channels*base+k]))) : (startval[k] + part*(vol*A.ptr[k] - startval[k])));
						
					}
					
					for(unsigned char k = 0; k < A.channels; ++k) startval[k] = B[A.channels*b+k-A.channels];
				}
			}
		}
	}
}


void fullsample_hqbuff2(audio_buffer& A, int* B, unsigned int b, double vol, int* startval)
{
	if(A.fill*b != 0) //should be some kind of error... or not?
	{
		//cout << "fullsample got inbuffer at " << A << " (" << a << " samples) and outbuffer at " << B << " (" << b << " samples)" << endl;
		if(vol == 0)
		{
			for(unsigned int i = 0; i < 2*b; ++i) B[i] = 0;
		}
		else
		{
			if(A.fill == b) for(unsigned int i = 0; i < A.channels*b; ++i) B[i] = (int) (vol*A.ptr[i]);
			else
			{
				float r = 1.0*A.fill/b;
				switch(A.channels)
				{
					case 2:
						if(r > 1) //median
						{
							for(unsigned int i = 0; i < b; ++i)
							{
								float sum = 0;
								float sum2 = 0;
								unsigned int n = 0;
								for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
								{
									++n;
									sum += A.ptr[2*j];
									sum2 += A.ptr[2*j+1];
								}
								B[2*i] = (int) (vol*sum/n); //gerade
								B[2*i+1] = (int) (vol*sum2/n); //ungerade
							}
						}
						else //(a < b)... interpolate
						{
							
							for(unsigned int i = 1; i <= b; ++i)
							{
								//i-1 is the sample index!!!
								float part = r*i;  //0.67 ... 1 means the first "present" input sample (A[0] and A[1])
								int base = (int) part; // -1: history, 0 - A[0]
								part = part - base; // tending with 0.67 to 1 (the first input sample)
								--base;
								B[2*(i-1)] = (int) ((base > -1) ? (vol*(A.ptr[2*base] + part*(A.ptr[2*(base+1)] - A.ptr[2*base]))) : (startval[0] + part*(vol*A.ptr[0] - startval[0]))); //gerade
								B[2*i-1] = (int) ((base > -1) ? (vol*(A.ptr[2*base+1] + part*(A.ptr[2*(base+1)+1] - A.ptr[2*base+1]))) : (startval[1] + part*(vol*A.ptr[1] - startval[1]))); //ungerade
							}
							startval[0] = B[2*b-2]; startval[1] = B[2*b-1];
						}
					break;
					case 1:
						if(r > 1) //median
						{
							for(unsigned int i = 0; i < b; ++i)
							{
								float sum = 0;
								unsigned int n = 0;
								for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
								{
									++n;
									sum += A.ptr[j];
								}
								B[i] = (int) (vol*sum/n); 
							}
						}
						else //(a < b)... interpolate
						{
							
							for(unsigned int i = 1; i <= b; ++i)
							{
								//i-1 is the sample index!!!
								float part = r*i;  //0.67 ... 1 means the first "present" input sample (A[0] and A[1])
								int base = (int) part; // -1: history, 0 - A[0]
								part = part - base; // tending with 0.67 to 1 (the first input sample)
								--base;
								
								B[(i-1)] = (int) ((base > -1) ? (vol*(A.ptr[base] + part*(A.ptr[base+1] - A.ptr[base]))) : (startval[0] + part*(vol*A.ptr[0] - startval[0])));
								
							}
							
							startval[0] = B[b-1];
						}
					break;
					default:
						if(r > 1) //median
						{
							float* sum = new float[A.channels];
							for(unsigned int i = 0; i < b; ++i)
							{
								for(unsigned char k = 0; k < A.channels; ++k) sum[k] += 0;
								unsigned int n = 0;
								for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
								{
									++n;
									for(unsigned char k = 0; k < A.channels; ++k) sum[k] += A.ptr[A.channels*j+k];
								}
								for(unsigned char k = 0; k < A.channels; ++k) B[A.channels*i+k] = (int) (vol*sum[k]/n); 
							}
							delete sum;
						}
						else //(a < b)... interpolate
						{
							
							for(unsigned int i = 1; i <= b; ++i)
							{
								//i-1 is the sample index!!!
								float part = r*i;  //0.67 ... 1 means the first "present" input sample (A[0] and A[1])
								int base = (int) part; // -1: history, 0 - A[0]
								part = part - base; // tending with 0.67 to 1 (the first input sample)
								--base;
								
								for(unsigned char k = 0; k < A.channels; ++k) B[A.channels*(i-1)+k] = (int) ((base > -1) ? (vol*(A.ptr[A.channels*base+k] + part*(A.ptr[A.channels*(base+1)+k] - A.ptr[A.channels*base+k]))) : (startval[k] + part*(vol*A.ptr[k] - startval[k])));
								
							}
							
							for(unsigned char k = 0; k < A.channels; ++k) startval[k] = B[A.channels*b+k-A.channels];
						}
				}
			}
		}
	}
}
#endif

/*
	resample A.fill samples from buffer A to b samples in appended to buffer B
	with channel conversion included...
	
	only for 2on2, 1on1, 2on1, 1on2 and XonX (not general XonY) channel situations
	
	I spare me the vol=0 optimization since this dumb situation doesn't seem to be worth it
	
	there are always two startvals saved (possibly identical)
*/
void fullchansample(audio_buffer& A, mixer_buffer& B, unsigned int b, double vol, int* startval)
{
//  bad check with restricted data type... 4096**3 does make zero...
//	if(A.fill*b*B.size != 0) //should be some kind of error... or not?
//	{
	if(b) //at least no division by zero
	{
		int* bpoint = B.ptr + B.fill*B.channels; //b should only be filled up
		float r = 1.0*A.fill/b; //resampling factor
		
		if(A.channels == B.channels)
		{
			sxdebug("fullchansample evenmix");
			if(A.fill == b) for(unsigned int i = 0; i < A.channels*b; ++i) bpoint[i] = (int) (vol*A.ptr[i]);
			else
			{
				if(A.channels == 2)
				{
					//This is the proven code of fullsample_hq for stereo.
					if(r > 1) //a > b ... downsample, arithmetic mean
					{
						sxdebug("fullchansample 2on2 downsample");
						for(unsigned int i = 0; i < b; ++i) //samples of B
						{
							float sum = 0;
							float sum2 = 0;
							unsigned int n = 0;
							for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
							{
								++n;
								sum += A.ptr[2*j];
								sum2 += A.ptr[2*j+1];
							}
							bpoint[2*i] = (int) (vol*sum/n); //gerade
							bpoint[2*i+1] = (int) (vol*sum2/n); //ungerade
						}
					}
					else //(a < b)... linear interpolation
					{
						sxdebug("fullchansample 2on2 upsample");
						for(unsigned int i = 1; i <= b; ++i) //samples of B
						{
							//i-1 is the sample index!!!
							float part = r*i;  //0.67 ... 1 means the first "present" input sample (A.ptr[0] and A.ptr[1])
							int base = (int) part; // -1: history, 0 - A.ptr[0]
							part = part - base; // tending with 0.67 to 1 (the first input sample)
							--base;
							bpoint[2*(i-1)] =
								(int)
								(
									(base > -1) ?
									(vol*(part*((int)A.ptr[2*(base+1)] - A.ptr[2*base]) + A.ptr[2*base])) :
									(startval[0] + part*(vol*A.ptr[0] - startval[0]))
								); //gerade
							bpoint[2*i-1] =
								(int)
								(
									(base > -1) ?
									(vol*(part*((int)A.ptr[2*(base+1)+1] - A.ptr[2*base+1]) + A.ptr[2*base+1])) :
									(part*(vol*A.ptr[1] - startval[1]) + startval[1])
								); //ungerade
						}
					}
					startval[0] = bpoint[2*b-2]; startval[1] = bpoint[2*b-1];
				}
				else //mono!!!!
				{
					if(r > 1) //a > b ... downsample, arithmetic mean
					{
						sxdebug("fullchansample 1on1 downsample");
						for(unsigned int i = 0; i < b; ++i) //samples of B
						{
							float sum = 0;
							unsigned int n = 0;
							for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
							{
								++n;
								sum += A.ptr[j];
							}
							bpoint[i] = (int) (vol*sum/n);
						}
					}
					else //(a < b)... linear interpolation
					{
						sxdebug("fullchansample 1on1 upsample");
						for(unsigned int i = 1; i <= b; ++i) //samples of B
						{
							//i-1 is the sample index!!!
							float part = r*i;  //0.67 ... 1 means the first "present" input sample (A.ptr[0] and A.ptr[1])
							int base = (int) part; // -1: history, 0 - A.ptr[0]
							part = part - base; // tending with 0.67 to 1 (the first input sample)
							--base;
							bpoint[i-1] =
								(int)
								(
									(base > -1) ?
									(vol*((int)A.ptr[base] + part*(A.ptr[base+1] - A.ptr[base]))) :
									(part*((int)vol*A.ptr[0] - startval[0]) + startval[0])
								);
						}
					}
					startval[0] = startval[1] = bpoint[b-1];
				}
			}
		}
		else
		{
			if(A.channels == 1) //upmix
			{
				sxdebug("fullchansample upmix");
				if(A.fill == b) for(unsigned int i = 0; i < A.fill; ++i) bpoint[2*i] = bpoint[2*i+1] = (int) (vol*A.ptr[i]);
				else
				{
					//This is the proven code of fullsample_hq for stereo.
					if(r > 1) //a > b ... downsample, arithmetic mean
					{
						sxdebug("fullchansample 1on2 downsample");
						for(unsigned int i = 0; i < b; ++i) //samples of B
						{
							float sum = 0;
							unsigned int n = 0;
							for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
							{
								++n;
								sum += A.ptr[j];
							}
							bpoint[2*i] = bpoint[2*i+1] = (int) (vol*sum/n); //gerade
						}
					}
					else //(a < b)... linear interpolation
					{
						sxdebug("fullchansample 1on2 upsample");
						for(unsigned int i = 1; i <= b; ++i) //samples of B
						{
							//i-1 is the sample index!!!
							float part = r*i;  //0.67 ... 1 means the first "present" input sample (A.ptr[0] and A.ptr[1])
							int base = (int) part; // -1: history, 0 - A.ptr[0]
							part = part - base; // tending with 0.67 to 1 (the first input sample)
							--base;
							bpoint[2*(i-1)] = bpoint[2*i-1] =
								(int)
								(
									(base > -1) ?
									(vol*(part*((int)A.ptr[base+1] - A.ptr[base]) + A.ptr[base])) :
									(part*(vol*A.ptr[0] - startval[0]) + startval[0])
								);
						}
					}
					startval[0] = startval[1] = bpoint[2*b-2];
				}
			}
			else //downmix
			{
				sxdebug("fullchansample downmix");
				if(A.fill == b) for(unsigned int i = 0; i < b; ++i) bpoint[i] = (int) (vol*( (int) A.ptr[2*i] + A.ptr[2*i+1] ));
				else
				{
					if(r > 1) //a > b ... downsample, arithmetic mean
					{
						sxdebug("fullchansample 2on1 downsample");
						for(unsigned int i = 0; i < b; ++i) //samples of B
						{
							float sum = 0;
							unsigned int n = 0;
							for(unsigned int j = (unsigned int) (r*i); j < (unsigned int) (r*(i+1)); ++j)
							{
								n += 2;
								sum += A.ptr[2*j] + A.ptr[2*j+1];
							}
							bpoint[i] = (int) (vol*sum/n);
						}
					}
					else //(a < b)... linear interpolation
					{
						sxdebug("fullchansample 2on1 upsample");
						for(unsigned int i = 1; i <= b; ++i) //samples of B
						{
							//i-1 is the sample index!!!
							float part = r*i;  //0.67 ... 1 means the first "present" input sample (A.ptr[0] and A.ptr[1])
							int base = (int) part; // -1: history, 0 - A.ptr[0]
							part = part - base; // tending with 0.67 to 1 (the first input sample)
							--base;
							bpoint[i-1] =
								(int)
								(
									(base > -1) ?
									(vol*(part*((int)A.ptr[2*(base+1)] - A.ptr[2*base] + A.ptr[2*(base+1)+1] - A.ptr[2*base+1]) + A.ptr[2*base] + A.ptr[2*base+1])/2) :
									( part*(vol*A.ptr[0] - startval[0] + vol*A.ptr[1] - startval[1]) + startval[0] +  startval[1])/2
								);
						}
					}
					startval[0] = startval[1] = bpoint[b-1];
				}
			}
		}
		B.fill += b; //we have done something wrong if this isn't the case
	}
//	else cerr << "[fullchansample] invalid call with A.fill=" << A.fill << " b=" << b << " B.size=" << B.size << endl;
	sxdebug("fullchansample done");
}


