/*
	in_sine: sine tone generator input device
	
	part of DerMixD
	(c)2006 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "audio_functions.hxx"
#include "input.hxx"
#include "in_sine.hxx"
#include "debug.hxx"

#include <math.h>

#define ME "sine_input"
const string sine_input::t = "sine";

sine_input::sine_input(input_data* ind)
{
	construct(ME);
	type = &t;
	data = ind;
	frequency = 440;
	data->format.rate = 44100;
	data->format.channels = 1;
}

sine_input::~sine_input()
{
	destruct(ME);
}

int sine_input::load(comm_data* cd, const string& filename) 
{
	debug("loading resource %s", filename.c_str());
	data->position = 0;
	data->at_end = false;
	data->filename = filename;
	// sine://<freq>@<sampling>
	if(filename.length() > 7)
	{
		string::size_type i;
		i = filename.find("@", 7);
		if(i != string::npos)
		{
			if(i > 7) frequency = atoi(filename.substr(7, i-7).c_str());
			long temp = atol(filename.substr(i+1, filename.length()-i).c_str());
			if(temp > 0) data->format.rate = temp;
		}
		else frequency = atoi(filename.c_str()+7);
		if(frequency < 0) frequency *= -1;
		data->length = data->format.rate; //1s... need a special "unspecified" value here...
		data->format.channels = 1;
		debug("have new settings: %dHz tone sampled with %luHz", frequency, data->format.rate);
		if(build_table()) cd->errcode = 0;
		else
		{
			cd->errcode = 124;
			cd->errtext = "error building table";
		}
	}
	else
	{
		cd->errcode = 123;
		cd->errtext = "no good sine pseudo-URL";
	}
	sem_post(cd->waitress);
	return 0;
}

// one should consider periodicity of sine
// sin(2pi * freq * pos/rate) == sin(2pi * (freq*pos/rate-floor(freq*pos/rate)) )
// with rate discrete values of the fractional term
// so, a table 0..rate-1, lookup index: pos % rate
bool sine_input::build_table()
{
	if(table.reallocate(1, data->format.rate) == data->format.rate)
	{
		sdebug("filling allocated table");
		for(unsigned int i = 0; i < data->format.rate; ++i)
		table.ptr[i] = (audio_type) (20000.0 * sin(2.0*3.14159265358979323846 * frequency * i / data->format.rate));
		sdebug("done filling table");
		return true;
	}
	else return false;
}

// only error that can occur is missing / wrong table... I suppose a successful load prevents this
// and without load you shall not play...
int sine_input::play(unsigned int n_samples)
{
	debug("playing with %u channels", data->format.channels);
	if(n_samples > buffer.size)
	{
		if( buffer.reallocate(data->format.channels, n_samples) == 0 ){  cout << "error on reallocating memory!\n"; return -1; }
	}
	for(unsigned int i = 0; i < n_samples; ++i)
	buffer.ptr[i] = table.ptr[(data->position+i) % data->format.rate];
	sdebug("filled");
	buffer.fill = n_samples;
	data->position += n_samples;
	sem_post(data->insem);
	sdebug("posted");
	return 0;
}

int sine_input::sampleseek(comm_data* cd, unsigned long position)
{
	data->position = position;
	data->at_end = false;
	
	if(cd != NULL)
	{
		cd->errcode = 0;
		cd->retval = seconds(&(data->format), position);
		sem_post(cd->waitress);
	}
	return 0;
}

int sine_input::seek(comm_data* cd, double position) //seek to sample
{
	unsigned long samplepos = samples(&(data->format), position);
	return sampleseek(cd, samplepos);
}

//need real eq...
int sine_input::set_eq(comm_data* cd)
{
	if(cd != NULL) sem_post(cd->waitress);
	return 0;
}
