// input_base: default stuff for input devices

/*
	input_base: base implementation of input_device class

	There is a worker thread and methods to trigger actions.
	What's mainly missing is the input_file implementation that's used by the worker.
	That you provide in the special class...

	part of DerMixD
	(c)2007-8 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "input.hxx"
#include "input_actions.hxx"
#include "action_errors.hxx"
#include "audio_functions.hxx"
#include "comparse.hxx"
#include "scanner.hxx"
#include "mixer.hxx"

#define ME "input_device"
static const string t = "input";

// Need to define that here to avoid the  undefined reference to `vtable for input_file'.
input_file::~input_file(){}

void input_file::explain_error(comm_data *cd, unsigned char aftercode, const string aftertext)
{
	if(cd == NULL) return;

	cd->errcode = aftercode;
	cd->errtext = aftertext;
	if(errcode != 0)
	{
		char num[4];
		snprintf(num, 4, "%i", (int)errcode);
		num[3] = 0;
		cd->errtext += "<- " + errtext + "(" + num + ")";
	}
}

// Transitional code.
bool input_file::do_seek(input_data *data)
{
	return false;
}

bool input_file::do_seek(input_data *data, unsigned long pos)
{
	data->position = pos;
	return do_seek(data);
}

void* worker(void *p)
{
	inwork_data *w = (inwork_data*) p;
	input_data *d = w->data;
	audio_buffer *b = w->buffer;
	input_file *i = w->inf;
	debug("[input worker %i] startup", d->id);
	w->thread_workout(1);
	while(1)
	{
		action *act; // current action
		int aid; // the action's ID
		// Wait for signal, then pull one action from work queue and process it.
		sxdebug("waiting");
		act = w->worklist.pull(); // That does the waiting, too.
		if(act == NULL)
		{
			serror("Pulled a null action!");
			continue;
		}
		aid = act->def->id;
		xdebug("processing action %i", aid);
		switch(aid)
		{
			case INWORK_PLAY:
			if(w->new_pos) /* Do I really still want that feature? */
			{
				d->at_end  = false; // If seek hits the end again, we'll know later.
				w->new_pos = false;
				if(!i->do_seek(d, w->the_new_pos))
				{
					error("[readily %i] ERROR there!", d->id);
					sem_post(d->insem);
					break;
				}
			}
			{
				unsigned long wanted_samples = (unsigned long) act->ints[0];
				b->fill = 0; // Fill it from beginning.
				// There's no comm_data for gimmes...
				if(    (wanted_samples > b->size || d->format.channels > b->channels)
				    && (b->reallocate(d->format.channels, wanted_samples) == 0)       )
				error("[inch %i] unable to resize buffer!", d->id);
				else if(!(i->do_read(b, wanted_samples)))
				error("[inch %i] unable to read from input file!", d->id);

				d->position += b->fill;
				if(b->fill < wanted_samples)
				{
					d->at_end = true;
					d->length = d->position;
				}
				else
				if(d->position >= d->length) d->length = d->position;

				sem_post(d->insem); // Always posting the same playback semaphore instead of one in the action...
			}
			break;
			case INWORK_OPEN:
			{
				debug("opening with (i=%p)", i);
				d->filename = act->strings[0]; // Lock that, too?
				if(i->do_open(d) && b->reallocate(d->format.channels, param.as_uint("buffer")))
				{
					d->position = 0;
					d->at_end = false;
				}
				else if(act->comdat != NULL) i->explain_error(act->comdat, 1, "open failed");
				else error("[inch %i] could not open %s and got noone to tell about it", d->id, d->filename.c_str());
			}
			break;
			case INWORK_SEEK:
			{
				d->at_end  = false; // If seek hits the end again, we'll know later.
				if(i->do_seek(d, (unsigned long)act->ints[0]))
				{
					if(act->comdat != NULL) act->comdat->retval = seconds(&(d->format), d->position);
				}
				else if(act->comdat != NULL) i->explain_error(act->comdat, 4, "seek did not work");
				else error("[inch %i] could not seek and got noone to tell about it", d->id);
			}
			break;
			case INWORK_PRESEEK:
				w->the_new_pos = (unsigned long)act->ints[0];
				w->new_pos     = true; //seek-before-play
			break;
			case INWORK_QUIT:
				i->do_close();
			break;
			case INWORK_LENGTH:
				if(i->do_length(d))
				{
					if(act->comdat != NULL)
					{
						act->comdat->errcode = 0;
						act->comdat->retval = seconds(&d->format, d->length);
						act->comdat->retval_ulong = d->length;
						act->comdat->retval_ulong2 = d->format.rate;
					}
				}
				else
				if(act->comdat != NULL) i->explain_error(act->comdat, NOT_SUPPORTED, "function not implemented?");
			break;
			case INWORK_EQ:
				if(!i->do_eq(act->floats))
				if(act->comdat != NULL) i->explain_error(act->comdat, NOT_SUPPORTED, "no equalizer implemented (should be non-fatal!)?");
			break;
			case INWORK_SCAN:
			{
				vector<string> scans;
				vector<scanner*> scanners;
				// Determine the scanners and get them up.
				split(act->strings[0], scans);
				for(vector<string>::iterator si=scans.begin(); si!=scans.end(); ++si)
				scanners.push_back(create_scanner(*si, &d->format));
				// Seek to beginning.
				unsigned long oldpos = d->position;
				if(!i->do_seek(d, 0))
				{
					if(act->comdat != NULL) i->explain_error(act->comdat, 4, "seek-0 did not work");
					else error("[inch %i] could not seek-0 and got noone to tell about it", d->id);

					break;
				}
				// Then slurp it all through a private buffer and scan.
				audio_buffer scanb;
				// Don't be too greedy with I/O here, trying same buffer size as mixer...
				if(scanb.reallocate(d->format.channels, buffer_samples) == 0)
				{
					serror("SCAN BUFFER!");
					if(act->comdat != NULL) i->explain_error(act->comdat, 6, "couldn't allocate scan buffer");
					else error("[inch %i] could not allocate and got noone to tell about it", d->id);

					break;
				}
				while(1)
				{
					// Try to get the lock once here... actually just to block when mixer wants to get the important data.
					// This way the other input threads, sharing the same priority, get a chance to do their work uninterrupted.
					// It is a hard measure here, introducing all these mutex operations... but it seems to work!
					pthread_mutex_lock(&playlock);
					pthread_mutex_unlock(&playlock);
					if(!i->do_read(&scanb, scanb.size))
					{
						if(act->comdat != NULL) i->explain_error(act->comdat, 6, "failed to read");
						else error("[inch %i] failed to read and got noone to tell about it", d->id);

						break;
					}
					for(vector<scanner*>::iterator i=scanners.begin(); i!=scanners.end(); ++i)
					if(*i != NULL) (*i)->scan(scanb);

					if(scanb.fill < scanb.size) break;

					scanb.fill = 0;
				}

				if(!i->do_seek(d, oldpos))
				{
					if(act->comdat != NULL) i->explain_error(act->comdat, 4, "seek-back did not work");
					else error("[inch %i] could not seek-back and got noone to tell about it", d->id);
					// No break... retlines could still be useful... but make sure they're deleted!
				}

				for(vector<scanner*>::iterator i=scanners.begin(); i!=scanners.end(); ++i)
				{
					act->comdat->retlines.push_back(new tstring);
					if(*i != NULL)
					{
						(*i)->report(*act->comdat->retlines.back());
						(*i)->update_input(d);
					}
					else *act->comdat->retlines.back() = "error: no scanner allocated";
				}
			}
			break;
			default: error("[inch %i] What to do for action %i?", d->id, aid);
		}
		if(act->comdat != NULL) sem_post(act->comdat->waitress);

		delete act; // Done with the action, delete it.
		if(aid == INWORK_QUIT) break;
	}
	debug("[input worker %i] leaving", d->id);
	pthread_exit(NULL);
}

input_device::input_device()
{
	construct(ME);
	on_line = false;
	wd.buffer = &buffer;
	wd.new_pos = false;
	wd.the_new_pos = 0;
	type = &t;
}

input_device::~input_device()
{
	destbegin(ME);
	if(on_line)
	{ // This is a programming error. Fix the derived input device class' destructor.
		serror("Whoa! I am still on line? didn't you call destroy()? Going to bomb out.");
		destroy();
	}
	destend(ME);
}

void input_device::build(input_data *ind, input_file *inf)
{
	data = ind;
	wd.inf = inf;
	wd.data = data;
	if(pthread_create(&wt, NULL, worker, (void*) &wd) == 0)
	{
		add_thread(wt, *type+" worker", &wd);
		on_line = true;
	}
	else problem = "cannot create worker thread";
}

void input_device::destroy()
{
	if(on_line)
	{
		action *chore = new action(NULL, &inwork_api[INWORK_QUIT]);
		chore->comdat = NULL;
		wd.worklist.push(chore);
		pthread_join(wt, NULL);
		on_line = false;
	}
}

int input_device::load(comm_data* cd, const string& filename) 
{
	action *chore = new action(NULL, &inwork_api[INWORK_OPEN]);
	debug("[inch %i] loading %s", data->id, filename.c_str());
	chore->strings.push_back(filename);
	chore->comdat = cd;
	wd.worklist.push(chore);
	return 0;
}

int input_device::play(unsigned int n_samples)
{
	action *chore = new action(NULL, &inwork_api[INWORK_PLAY]);
	chore->ints.push_back(n_samples);
	chore->comdat = NULL;
	wd.worklist.push(chore);
	return 0;
}

int input_device::sampleseek(comm_data* cd, unsigned long position)
{
	action *chore = new action(NULL, &inwork_api[cd != NULL ? INWORK_SEEK : INWORK_PRESEEK]);
	debug("[inch %i] seeking to %lu", data->id, position);
	chore->ints.push_back((long)position);
	chore->comdat = cd;
	wd.worklist.push(chore);
	return 0;
}

void input_device::scan_length(comm_data *cd)
{
	sdebug("default scan_length");
	if(cd != NULL)
	{
		action *chore = new action(NULL, &inwork_api[INWORK_LENGTH]);
		chore->comdat = cd;
		wd.worklist.push(chore);
	}
}

void input_device::no_scan_length(comm_data *cd)
{
	sdebug("no_scan_length");
	if(cd != NULL)
	{
		debug("sending an error via %p", (void*)cd);
		cd->errcode = NOT_SUPPORTED;
		cd->errtext = "function not implemented";
		sem_post(cd->waitress);
	}
}

void input_device::done_scan_length(comm_data *cd)
{
	sdebug("done_scan_length");
	if(cd != NULL)
	{
		cd->errcode = 0;
		cd->retval = seconds(&data->format, data->length);
		cd->retval_ulong = data->length;
		cd->retval_ulong2 = data->format.rate;
		sem_post(cd->waitress);
	}
}

void input_device::scan(comm_data *cd, const string &list)
{
	sdebug("scan");
	if(cd != NULL)
	{
		action *chore = new action(NULL, &inwork_api[INWORK_SCAN]);
		chore->comdat = cd;
		chore->strings.push_back(list);
		wd.worklist.push(chore);
	}
}

int input_device::seek(comm_data* cd, double position)
{
	unsigned long samplepos = samples(&(data->format), position);
	return sampleseek(cd, samplepos);
}

bool input_device::online()
{
	return on_line;
}

bool input_device::ready()
{
	return on_line;
}

//need real eq...
int input_device::set_eq(comm_data* cd)
{
	action *chore = new action(NULL, &inwork_api[INWORK_EQ]);
	chore->comdat = cd;
	// I cannot push references to volatile double... either this copy or const_cast...
	double tmp = data->bass;
	chore->floats.push_back(tmp);
	tmp = data->mid;
	chore->floats.push_back(tmp);
	tmp = data->treble;
	chore->floats.push_back(tmp);
	wd.worklist.push(chore);
	return 0;
}
