/*
	netcomm: tcp port watching and basic socket communication
	
	part of DerMixD
	(c)2004/5 Thomas Orgis, licensed under GPLv2
	
	derived from:
	
	netcomm.c for mixplayd
	
	provides functions for communiction on a tcp socket
	(c) Siegfried Wagner 01/06/27
	
	license: GNU General Public License (GPL)
	         For details see the file COPYING included with this archive
	last update: 01/08/28 by Siegfried Wagner
*/

#include <fcntl.h>
#include <sstream>
#include <sys/un.h>
#include "common.hxx"
#include "debug.hxx"
#include "netcomm.hxx"


//MSG_NOSIGNAL doesn't exist on Tru64...
//but it is _very_ handy when the client already closed the connection and I want to write, say, just \r\n
//without nosignal, dermixd dies...
//so, let's just ignore SIGPIPE in the main thread and additionally define it to zero here if MSG_NOSIGNAL is not there...
//I guess that even when ignoring the signal in won't hurt preventing it in recv/send.
//Beware: dermixd even dies on careless send(..., "\r\n", ) if the client is away regardless of MSG_NOSIGNAL or ignoring - but only on slow machines or otherwise slowed down (inside gdb)!
#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

// Create a TCP socket.
int bind_tcp(porttype port_number, int local);
// Create a UNIX domain socket in file system.
int bind_unix(const char *name);


socket_listener::socket_listener(void)
{
	construct("socket_listener");
	online = false;
}

socket_listener::~socket_listener()
{
	destruct("socket_listener");
}

port_listener::port_listener(porttype port_number, int local)
{
	construct("port_listener");
	sock = bind_tcp(port_number, local);
	online = (sock < 0) ? false : true;
}

unix_listener::unix_listener(const string name)
{
	construct("unix_listener");
	sockname = name;
	sock = bind_unix(name.c_str());
	online = (sock < 0) ? false : true;
}

// Just close the TCP port.
port_listener::~port_listener()
{
	destbegin("port_listener");
	if(online){ close(sock); }
	destend("port_listener");
}

// Close the socket and remove the stale file system entry.
unix_listener::~unix_listener()
{
	destbegin("unix_listener");
	if(online)
	{
		close(sock);
		unlink(sockname.c_str());
	}
	destend("unix_listener");
}

int bind_tcp(porttype port_number, int local)
{
	int sock;
	struct sockaddr_in svr;
	int flag=1;
	
	bzero((char*) &svr, sizeof(svr)); //important on Tru64/Alpha!
	svr.sin_family = AF_INET;
	svr.sin_port = htons(port_number);
	if(local) svr.sin_addr.s_addr = inet_addr("127.0.0.1");
	else svr.sin_addr.s_addr = INADDR_ANY;
	
	if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror ("error creating listening socket");
		return -1;
	}
	
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag)) < 0)
	{
		perror ("error setting SO_REUSEADDR on listening socket");
		return -1;
	}
	
	if(bind(sock, (struct sockaddr *) &svr, sizeof(svr)) < 0)
	{
		fprintf (stderr, "bind: failed to bind to port %d: %s.\n\n", port_number, strerror (errno));
		if(port_number < 1024 && getuid ())
			fprintf (stderr, "A likely reason is that you tried to bind low-numbered ports but you are not root!\n");
		close (sock);
		return -1;
	}
	
	if (listen (sock, SOMAXCONN) < 0)
	{
		fprintf (stderr, "error listening: %s\n", strerror (errno));
		close (sock);
		return -1;
	}
	
	//I want blocking
	//fcntl(sock, F_SETFL, O_NONBLOCK);
	fcntl(sock, F_SETFD, FD_CLOEXEC); //close on exec (after forking)
	
	return sock;
}

int bind_unix(const char *name)
{
	int sock;
	struct sockaddr_un svr;

	if(strlen(name) >= sizeof(svr.sun_path)/sizeof(char))
	{
		error("Too long UNIX domain socket name: %s", name);
		return -1;
	}
	bzero((char*) &svr, sizeof(svr)); //important on Tru64/Alpha?

	if( (sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
	{
		error("Unable to create listening UNIX domain socket (%s).", strerror(errno));
		return -1;
	}
	svr.sun_family = AF_UNIX;
	strcpy(svr.sun_path, name);
	unlink(name);
	if(bind(sock, (struct sockaddr *) &svr, sizeof(svr)) < 0)
	{
		error("Bind for UNIX domain socket \"%s\" failed: %s.", name, strerror(errno));
		close (sock);
		return -1;
	}
	if(listen (sock, SOMAXCONN) < 0)
	{
		error("No listening: %s", strerror(errno));
		close(sock);
		return -1;
	}
	fcntl(sock, F_SETFD, FD_CLOEXEC); //close on exec (after forking)

	return sock;
}

//wait for and accept connection requests (delivering new socket for communication)
int socket_listener::accept_connection()
{
	//struct sockaddr_in cli;
	//socklen_t clilen;
	int clisock;
	
	//clilen = (socklen_t) sizeof(cli);
	//clisock = accept(sock, (struct sockaddr *)&cli, &clilen);
	clisock = accept(sock, NULL, NULL);
	
	//I want blocking...
	//fcntl(clisock, F_SETFL, O_NONBLOCK);
	fcntl(clisock, F_SETFD, FD_CLOEXEC); //close on exec (after forking)
	
	return clisock;
}

//simple reader object


socket_buffer::socket_buffer(int s)
{
	construct("socket_buffer");
	sock = s;
	buf[MSG_BUF_SIZE-1] = '\0';
	pos = 0;
	fill = 0;
	{
		ostringstream os;
		os << "[spy " << sock;
		spyinprefix  = os.str() + " in] ";
		spyoutprefix = os.str() + " out] ";
	}
}

socket_buffer::~socket_buffer()
{
	destbegin("socket_buffer");
	close(sock);
	destend("socket_buffer");
}

/*
algo:

when still sth. in buffer (indicated by buffed == true, beginning at pos)
begin search for next \n
if found before end: make \0, add to msg, update the pos, 

this may go on in a loop if the line to read is really long... 

*/
int socket_buffer::read_line(string& msg)
{
	int msglen = 0;
	int i;
	
	while(1)
	{
		//cout << "fill: " << fill << endl;
		if(fill)
		{
			//search for next line end
			for(i=pos;i<fill;++i)
			{
				if(buf[i] == '\n'){ buf[i] = '\0'; break; }
				if(buf[i] == '\r'){ buf[i] = '\0'; }
			}
			//cout << "i: " << i << endl;
			if( (msglen += i - pos) > MAX_MSG_LENGTH )
			{
				cerr << "msglen of " << msglen << " is greater than " << MAX_MSG_LENGTH << endl;
				fill = 0;
				pos = 0;
				return -2; //interpet this!
			}
			//copy line starting at pos and ending at i
			msg += (buf + pos);
			if(i < fill) //really found a \n
			{
				//the line was the buffer: i is last character
				if(i == fill-1)
				{
					fill = 0;
					pos = 0;
				}
				//there is sth. left: i is smaller
				else
				{
					pos = i + 1;
				}
				return msglen;
			}
			//not at end: reading sth. next time...
			else{ fill = 0; pos = 0; } //pos=0 here or a few lines below?
		}
		else
		{
			pos = 0;
			#ifdef DEBUG_SOCKET
			cout << "[socket " << sock << "] recv" << endl;
			#endif
			fill = recv(sock, buf, MSG_BUF_SIZE-1, MSG_NOSIGNAL);
			#ifdef DEBUG_SOCKET
			cout << "[socket " << sock << "] got " << fill << " bytes" << endl;
			#endif
			if(fill < 1)
			{
				//we would have blocked if all was right, wouldn't we?
				//getting no data means that sth. is broken...
				//an error is an error anyway
				//if(fill < 0){
				 fill=0; msg = ""; return -1;
				//  }
				//handle this!!
				//return 0; //hm... read nothing... maybe having read sth. before
			}
		}
	}
}

int socket_buffer::write(string& msg)
{
	if (sock<0) return -1;
	#ifdef DEBUG_SOCKET
	cout << "[socket " << sock << "] going to write \"" << msg << "\"" << endl;
	#endif
	return send(sock, msg.c_str() , msg.length() ,  MSG_NOSIGNAL);
}

int socket_buffer::writeln(string* msg)
{
	if (sock<0) return -1;
	#ifdef DEBUG_SOCKET
	cout << "[socket " << sock << "] going to write \"" << *msg << "\"" << endl;
	#endif
	//Even with MSG_NOSIGNAL and signal(SIGPIPE,SIG_IGN), I get SIGPIPE raised and the program killed if I try to do the second send on the broken pipe!
	//Is it better to make just another copy of the string and modify it to append \r\n so that I can do a single send?
	if(send(sock, msg->c_str() , msg->length() ,  MSG_NOSIGNAL) != -1)
	{
		return send(sock, "\r\n", 2, MSG_NOSIGNAL);
	}
	else return -1;
}
