/*
	out_alsa: alsa output device
	
	part of DerMixD
	(c)2005 Thomas Orgis, licensed under GPLv2
*/

#include "common.hxx"
#include "audio_buffer.hxx"
#include "output.hxx"
#include "out_alsa.hxx"

//I don't like this!
const string alsa_output::t = "alsa";

//Underrun and suspend recovery; helper for writer
static int xrun_recovery(snd_pcm_t *handle, int err)
{
	if(err == -EPIPE)
	{
		/* under-run */
		err = snd_pcm_prepare(handle);
		if(err < 0) error("[alsa handle %p] Can't do recovery from underrun, prepare failed: %s", (void*)handle, snd_strerror(err));
		err = 0;
	}
	else if(err == -ESTRPIPE)
	{ 
		/* wait until the suspend flag is released */
		while((err = snd_pcm_resume(handle)) == -EAGAIN) sleep(1);
		if (err < 0)
		{
			err = snd_pcm_prepare(handle);
			if(err < 0) error("[alsa handle %p] Can't do recovery from suspend, prepare failed: %s", handle, snd_strerror(err));
		}
		err = 0;
	}
	return err;
}

void* alsa_writer(void* p)
{
	writerdata* data = (writerdata*) p;
	alsa_writerextra* extra = (alsa_writerextra*) data->extra;
	debug("[ch%i] writer thread started", data->id); 
	data->thread_workout(0);

	while(1)
	{
		safe_sem_wait(&data->insem);
		data->error = extra->err = 0;
		extra->ptr = const_cast<audio_type*> (data->buffer.ptr);
		extra->cptr = data->buffer.size;

		while (extra->cptr > 0)
		{
			xdebug("[ch%i] threaded write loop", data->id); 
			extra->err = snd_pcm_writei(extra->handle, extra->ptr, extra->cptr);
			if(extra->err == -EAGAIN) continue;
			if(extra->err < 0)
			{
				error("[ch%i] underrun!", data->id);
				if(xrun_recovery(extra->handle, extra->err) < 0)
				{
					error("[ch%i] write error: %s", data->id, snd_strerror(extra->err)); 
					data->error = extra->err;
				}
				break;  /* skip one period */
			}
			extra->ptr += extra->err * data->buffer.channels;
			extra->cptr -= extra->err;
    }
		sem_post(&data->outsem);
	}
	pthread_exit(NULL);
}

void alsa_output::special_write()
{
	wextra.ptr = const_cast<audio_type*> (wdat.buffer.ptr);
	wextra.cptr = wdat.buffer.size; // not better fill?

	while (wextra.cptr > 0)
	{
		xdebug("[ch%i] serial write loop", data->id); 

		wextra.err = snd_pcm_writei(wextra.handle, wextra.ptr, wextra.cptr);
		if(wextra.err == -EAGAIN) continue;
		if(wextra.err < 0)
		{
			error("[ch%i] underrun!", data->id);
			if(xrun_recovery(wextra.handle, wextra.err) < 0)
			{
				debug("[ch%i] write error: %s", data->id, snd_strerror(wextra.err)); 
				wdat.error = wextra.err;
			}
			break;  /* skip one period */
		}
		wextra.ptr += wextra.err * wdat.buffer.channels;
		wextra.cptr -= wextra.err;
	}
}

alsa_output::alsa_output(output_data* odi, unsigned int bufsize, const string type)
{
	construct("alsa_output");
	current_type = type;
	if(type == "alsa_serial") build(odi, bufsize, NULL);
	else if((type == "alsa") || (type == "alsa_threaded")) build(odi, bufsize, alsa_writer);
	else
	{
		problem = "unknown variant " + type + " ... I only know alsa, alsa_threaded (aliased by alsa) or alsa_serial";
		prebuild(); //destroy would do havoc trying to destroy nonexistent semahores and thread otherwise
	}
}

alsa_output::~alsa_output()
{
	destbegin("alsa_output");
	destroy();
	destend("alsa_output");
}

bool alsa_output::special_init()
{
	snd_pcm_hw_params_malloc(&hwparams);
 	stream = SND_PCM_STREAM_PLAYBACK;
	wdat.extra = (void*) &wextra;
	return true;
}

void alsa_output::special_exit()
{
	snd_pcm_hw_params_free(hwparams);
}

bool alsa_output::special_activate()
{
	int errora = 0;
	bool succ = false;

	// I still hate this deep if... but flattening would add more code...
	if((errora = snd_pcm_open(&pcm_handle, data->filename.c_str(), stream, 0)) > -1)
	{
		debug("[ch%i] opened %s", data->id, data->filename.c_str());
		if((errora = snd_pcm_hw_params_any(pcm_handle, hwparams)) > -1)
		{
			debug("[ch%i] params initialized", data->id);
			if((errora = snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) > -1)
			{
				debug("[ch%i] access set", data->id);
				if((errora = snd_pcm_hw_params_set_format(pcm_handle, hwparams, SND_PCM_FORMAT_S16)) > -1)
				{
					debug("[ch%i] format set", data->id);
					if((errora = snd_pcm_hw_params_set_rate(pcm_handle, hwparams, data->format.rate, 0)) > -1)
					{
						debug("[ch%i] rate set", data->id);
						if((errora = snd_pcm_hw_params_set_channels(pcm_handle, hwparams, (unsigned int)data->format.channels)) > -1)
						{
							debug("[ch%i] channels set", data->id);
							snd_pcm_uframes_t bufsize = (snd_pcm_uframes_t) (param.as_double("alsa","buffer")*data->format.rate);
							debug("[ch%i] bufsize: %lu", data->id, (unsigned long) bufsize);
							if((errora = snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, &bufsize)) > -1)
							{
								debug("[ch%i] buffer size set", data->id);
								//periods mean n-buffering? 
								snd_pcm_uframes_t perisize = bufsize / 4;
								if ((errora = snd_pcm_hw_params_set_period_size_near(pcm_handle, hwparams, &perisize, NULL)) > -1)
								{
									debug("[ch%i] periods set", data->id);
									if((errora = snd_pcm_hw_params(pcm_handle, hwparams)) > -1)
									{
										debug("[ch%i] params set", data->id);
										if((errora = snd_pcm_prepare(pcm_handle)) > -1) //&& ((errora = snd_pcm_start(pcm_handle)) > -1))
										{
											wextra.handle = pcm_handle;
											succ = true;
										}
										else problem = "prepare failed";
									}
									else problem = "params setting failed";
								}
								else problem = "periods setting failed";
							}
							else problem = "buffer size setting failed";
						}
						else problem = "channel setting failed";
					}
					else problem = "rate setting failed";
				}
				else problem = "format setting failed";
			}
			else problem = "access setting failed";
		}
		else problem = "param init failed";
	}
	else problem = "device open failed";
	debug("[ch%i] activate done; error value %i", data->id, errora);

	return succ;
}

void alsa_output::special_deactivate()
{
	debug("[ch%i] deactivating", data->id);
	snd_pcm_drain(pcm_handle);
	snd_pcm_close(pcm_handle);
}

//there is going to be some silence; make sure the last samples are played
void alsa_output::special_pause()
{
	debug("[ch%i] pause", data->id);
	snd_pcm_pause(pcm_handle,1);
}

